defmodule SqlMembershipProvider.Membership do
  @moduledoc """
  Struct for representing a user's membership to an application and their authentication information.
  """

  import SqlMembershipProvider.TdsUtils

  defstruct [
    :application_id,
    :user_id,
    :password,
    :password_format,
    :password_salt,
    :mobile_pin,
    :email,
    :lowered_email,
    :password_question,
    :password_answer,
    :is_approved,
    :is_locked_out,
    :create_date,
    :last_login_date,
    :last_password_changed_date,
    :last_lockout_date,
    :failed_password_attempt_count,
    :failed_password_attempt_window_start,
    :failed_password_answer_attempt_count,
    :failed_password_answer_attempt_window_start,
    :comment
  ]

  @type t() :: %SqlMembershipProvider.Membership{
          application_id: String.t(),
          user_id: String.t(),
          password: String.t(),
          password_format: integer(),
          password_salt: String.t(),
          mobile_pin: String.t() | nil,
          email: String.t() | nil,
          lowered_email: String.t() | nil,
          password_question: String.t() | nil,
          password_answer: String.t() | nil,
          is_approved: boolean(),
          is_locked_out: boolean(),
          create_date: tuple(),
          last_login_date: tuple(),
          last_password_changed_date: tuple(),
          last_lockout_date: tuple(),
          failed_password_attempt_count: integer(),
          failed_password_attempt_window_start: tuple(),
          failed_password_answer_attempt_count: integer(),
          failed_password_answer_attempt_window_start: tuple(),
          comment: String.t() | nil
        }

  @doc """
  Fetch a membership by user id.
  """
  @spec find_by_user_id(atom() | pid(), String.t()) :: SqlMembershipProvider.Membership.t() | nil
  def find_by_user_id(pid, user_id) when (is_atom(pid) or is_pid(pid)) and is_binary(user_id) do
    Tds.query!(
      pid,
      "SELECT TOP 1
        ApplicationId as application_id,
        UserId as user_id,
        Password as password,
        PasswordFormat as password_format,
        PasswordSalt as password_salt,
        MobilePIN as mobile_pin,
        Email as email,
        LoweredEmail as lowered_email,
        PasswordQuestion as password_question,
        PasswordAnswer as password_answer,
        IsApproved as is_approved,
        IsLockedOut as is_locked_out,
        CreateDate as create_date,
        LastLoginDate as last_login_date,
        LastPasswordChangedDate as last_password_changed_date,
        LastLockoutDate as last_lockout_date,
        FailedPasswordAttemptCount as failed_password_attempt_count,
        FailedPasswordAttemptWindowStart as failed_password_attempt_window_start,
        FailedPasswordAnswerAttemptCount as failed_password_answer_attempt_count,
        FailedPasswordAnswerAttemptWindowStart as failed_password_answer_attempt_window_start,
        Comment as comment
        FROM aspnet_Membership
        WHERE UserId = @user_id",
      [%Tds.Parameter{name: "@user_id", value: user_id}]
    )
    |> result_to_map()
    |> stringify_uuids(["application_id", "user_id"])
    |> to_struct(__MODULE__)
    |> List.first()
  end

  @doc """
  Check if a plaintext password matches a user's hashed password.
  """
  @spec is_password_valid?(
          SqlMembershipProvider.Membership.t(),
          String.t()
        ) :: boolean()
  def is_password_valid?(
        %SqlMembershipProvider.Membership{
          password: hashed_password,
          password_format: format,
          password_salt: salt
        },
        password
      )
      when is_binary(password) do
    hash_password(password, format, salt) == hashed_password
  end

  defp hash_password(password, 1, salt) do
    salt = Base.decode64!(salt)

    password =
      String.to_charlist(password)
      |> :unicode.characters_to_binary(:utf8, {:utf16, :little})

    :crypto.hash(:sha, salt <> password)
    |> Base.encode64()
  end

  defp hash_password(_password, format, _salt),
    do: raise(SqlMembershipProvider.UnsupportedPasswordFormatError, format)
end
