defmodule SqlMembershipProvider.Profile do
  @moduledoc """
  Struct for representing a user's profile information that may contain arbitrary properties
  """

  import SqlMembershipProvider.TdsUtils

  defstruct [
    :user_id,
    :property_names,
    :property_values_string,
    # FIXME: TDS throws error trying to retrieve an "image" type column
    # :property_values_binary,
    :last_updated_date
  ]

  @type t() :: %SqlMembershipProvider.Profile{
          user_id: String.t(),
          property_names: String.t(),
          property_values_string: String.t(),
          last_updated_date: tuple()
        }

  @doc """
  Fetch a profile by user id.
  """
  @spec find_by_user_id(atom() | pid(), String.t()) :: SqlMembershipProvider.Profile.t() | nil
  def find_by_user_id(pid, user_id) when (is_atom(pid) or is_pid(pid)) and is_binary(user_id) do
    Tds.query!(
      pid,
      "SELECT TOP 1
        UserId as user_id,
        PropertyNames as property_names,
        PropertyValuesString as property_values_string,
        LastUpdatedDate as last_updated_date
        FROM aspnet_Profile
        WHERE UserId = @user_id",
      [%Tds.Parameter{name: "@user_id", value: user_id}]
    )
    |> result_to_map()
    |> stringify_uuids(["user_id"])
    |> to_struct(__MODULE__)
    |> List.first()
  end

  @doc """
  Parse a profile's property fields into a map.
  """
  @spec properties(SqlMembershipProvider.Profile.t()) :: map()
  def properties(
        profile = %SqlMembershipProvider.Profile{
          property_names: property_names
        }
      ) do
    property_names
    |> String.split(":")
    |> Enum.chunk_every(4)
    |> Enum.map(&List.to_tuple/1)
    |> Enum.filter(fn
      {_key, _type, _start, _length} -> true
      _ -> false
    end)
    |> Enum.map(fn property -> extract_property(property, profile) end)
    |> Enum.into(%{})
  end

  defp extract_property({key, _type, _start, "-1"}, _profile) do
    {key, nil}
  end

  # properties found in property_values_binary
  defp extract_property({key, "B", _start, _length}, _profile) do
    # we probably can't deserialize the binary proerties
    {key, nil}
  end

  # properties found in property_values_string
  defp extract_property({key, "S", start, length}, %SqlMembershipProvider.Profile{
         property_values_string: property_values_string
       }) do
    start = String.to_integer(start)
    length = String.to_integer(length)

    value =
      String.codepoints(property_values_string)
      |> Enum.slice(start, length)
      |> Enum.join("")

    {key, value}
  end
end
