defmodule SqlMembershipProvider.TdsUtils do
  @moduledoc """
  Utility functions for converting `Tds.Result` into custom structs.

  Usually used like:
  ```
  result
  |> result_to_map()
  |> stringify_uuids(["user_id"])
  |> to_struct(__MODULE__)
  ```
  """

  @doc """
  Convert a `Tds.Result` into a list of maps.

  `Tds.Result` stores values as a list of rows containing a list of column values. Having this stored as a map lets you retrieve a column's value by it's column name.
  """
  @spec result_to_map(Tds.Result.t()) :: [map()]
  def result_to_map(%Tds.Result{columns: columns, rows: rows}) do
    rows
    |> Enum.map(fn row -> Enum.zip(columns, row) |> Map.new() end)
  end

  @doc """
  Convert UUID columns from Microsoft's mixed-endian format to little-endian format.
  """
  @spec stringify_uuids([map()], [String.t()]) :: [map()]
  def stringify_uuids(rows, columns_to_convert)
      when is_list(rows) and is_list(columns_to_convert) do
    Enum.map(rows, &stringify_uuids(&1, columns_to_convert))
  end

  @spec stringify_uuids(map(), [String.t()]) :: map()
  def stringify_uuids(row, columns_to_convert) when is_map(row) and is_list(columns_to_convert) do
    Enum.reduce(columns_to_convert, row, fn key, row ->
      Map.update!(row, key, &uuid_to_string!/1)
    end)
  end

  defp uuid_to_string!(nil), do: nil

  defp uuid_to_string!(uuid) do
    {:ok, uuid_str} = Tds.Types.UUID.load(uuid)
    uuid_str
  end

  @doc """
  Convert a map into a struct

  [Taken from José on elixir-lang-talk](https://groups.google.com/forum/#!msg/elixir-lang-talk/6geXOLUeIpI/L9einu4EEAAJ)
  """
  @spec to_struct(nil, atom()) :: nil
  def to_struct(nil, _kind), do: nil

  @spec to_struct([map()], module()) :: [struct()]
  def to_struct(rows, kind) when is_list(rows) do
    Enum.map(rows, &to_struct(&1, kind))
  end

  @spec to_struct(map(), module()) :: struct()
  def to_struct(attrs, kind) when is_map(attrs) do
    struct = struct(kind)

    Enum.reduce(Map.to_list(struct), struct, fn {key, _value}, acc ->
      case Map.fetch(attrs, Atom.to_string(key)) do
        {:ok, value} -> %{acc | key => value}
        :error -> acc
      end
    end)
  end
end
